#include "global.h"
#include "cartridge.h"
#include "mapper.h"
#include "file.h"


/* iNES header
BYTE	BIT	DESC							cartridge->var
0-3		NES^Z header
4		number of 16kB PRG banks				prgbanks (8k here)
5		number of 8kB VROM banks				vrombanks (1k here)
6	0	mirroring: 1=vert, 0=hor				mirroring (1100=h, 1010=v)
	1	battery-backed RAM: 1=true, 0=false			battery
	2	512B trainer at $7000-$71ff: 1=true, 0=false		trainer
	3	four-screen VRAM layout: 1=true, 0=false		mirroring|=10000000, unused if not MMC3
	4-7	lower nibble of ROM mappertype				type(converted)
7	0	VS-system: 1=true, 0=false				unused yet
	1-3	reserved, 0
	4-7	high nibble of ROM mappertype				type(converted)
8		number of 8kB (V)RAM? banks: 0=1, 1=1, 2=2, 3=3, etc.	mostly unused (only for sorom detection)
9	0	video mode: 1=PAL, 0=NTSC				mode, unused if not auto or byte 9>1/garbage
	1-7	reserved, 0
10-15		reserved, 0
16-...		(trainer)+PRG banks					trainer, prg[0..511]
...-end		VROM banks						vrom[0..2047] */

static void cartridge_check_mmc6(Cartridge* cart)
{
	/* check for mmc6 type by sav file size */
	file_setfile(&file->batterydir,cart->name,"sav",NULL);
	if (file_open()&&file->size==0x400) cart->type=MAPTYPE_NINTENDO_MMC6_HXROM;
	
	file_close();
}

int cartridge_load(int get,int* s_maptype,int* s_battery,int* s_mir,u32* s_crc,u32* s_crc_p)
{
	#define CL_ERROR() file_patch_close(); file_close(); return FALSE;
	
	u32 crc,crc_p=0;
	u8 header[0x10]={0};
	char fn[STRING_SIZE]={0};
	char n[STRING_SIZE]={0};
	int i=TRUE,dh=FALSE,mmc6check=FALSE,ines;
	Cartridge* cart=cartridge_temp;
	memset(cart,0,sizeof(Cartridge));
	
	/* open file, just to get checksum */
	if (!file_open()) { LOG(LOG_MISC|LOG_WARNING,"ROM open error\n"); CL_ERROR(); }
	if ((file->size<(0x2010))|(file->size>=CARTRIDGE_MAXSIZE)) { LOG(LOG_MISC|LOG_WARNING,"ROM read error\n"); CL_ERROR(); }
	crc=file->crc32;
	
	/* open optional ips */
	strcpy(fn,file->filename);
	strcpy(n,file->name);
	file_setfile(&file->patchdir,n,"ips",NULL);
	file_close();
	if (!file_open()) {
		file_close();
		file_setfile(&file->patchdir,n,"zip","ips");
		if (!file_open()) { i=FALSE; file_close(); }
	}
	crc_p=file->crc32;
	if (i&(crc_p!=crc)) if (!file_patch_init()) file_patch_close();
	file_close();
	
	/* really open file */
	file_setfile(NULL,fn,NULL,"nes");
	if (!file_open()) { LOG(LOG_MISC|LOG_WARNING,"ROM open error\n"); CL_ERROR(); }
	if (((file->size<(0x10+0x2000))|(file->size>=CARTRIDGE_MAXSIZE))||!file_read(header,0x10)) { LOG(LOG_MISC|LOG_WARNING,"ROM read error\n"); CL_ERROR(); }
	
	/* check for dirty header */
	if ((memcmp(header,"NES\x1a",4)!=0)|(header[7]&BIN8(00001110))|(header[9]>1)|(header[4]==0)|(memcmp(header+10,"\0\0\0\0\0\0",6)!=0)) {
		
		#define LWRONG_HEADER(x) LOG(LOG_MISC|LOG_WARNING,"%s%s\n",text,x)
		
		char text[STRING_SIZE]; char temp[STRING_SIZE];
		memset(temp,0,STRING_SIZE);
		for (i=0;i<0x10;i++) {
			temp[i]=(char)header[i];
			if (header[i]==0) temp[i]='.';
			else if ((header[i]==255)|(header[i]<32)) temp[i]=' ';
		}
		sprintf(text,"dirty header:[%s], ",temp);
		
		if (memcmp(header,"NES\x1a",4)!=0) { LWRONG_HEADER("no NES^Z"); dh|=1; } /* no NES^Z in header check */
		else if (header[4]==0) { LWRONG_HEADER("no PRG-ROM"); dh|=1; } /* 0 prg banks */
		else if (memcmp(header+7,"Dis",3)==0) { memset(header+7,0,9); LWRONG_HEADER("known"); } /* Dis(kDude!) (things like DisNi330 are also available) header check/clean: Mega Man (U), Jackal (U), Rush'n Attack (U), Startropics (U), Karate Champ (U), etc. */
		else if (memcmp(header+7,"demiforce",9)==0) { memset(header+7,0,9); LWRONG_HEADER("known"); } /* demiforce header check/clean: some translations by Neo Demiforce */
		else if (memcmp(header+9,"fortune",7)==0) { memset(header+9,0,7); LWRONG_HEADER("known"); } /* fortune header check/clean: Super Contra 7 (Unl), etc(?) */
		else if (memcmp(header+7,"byKH+KEN",8)==0) { memset(header+7,0,9); LWRONG_HEADER("known"); } /* byKH+KEN header check/clean: Hot Slot (UE), etc(?) */
		else if (memcmp(header+9,"by PIRO",7)==0) { memset(header+9,0,7); LWRONG_HEADER("known"); } /* by PIRO header check/clean: Takahashi Meijin no Bouken Shima IV (J), Momotarou Densetsu (J), etc. */
		else if (memcmp(header+8,"GitM",4)==0) { memset(header+8,0,8); LWRONG_HEADER("known"); } /* GitM* header check/clean: M.U.L.E. (U), etc(?) */
		
		/* something else: any bits set to 1 that should be 0 */
		else if ((header[7]&BIN8(00001110))|(header[8]>8)|(header[9]>1)) { LWRONG_HEADER("unknown"); memset(header+7,0,9); } /* most of these some garbage followed by "Ni330", remove everything starting from high mappernumber bits */
		else { LWRONG_HEADER("harmless"); } /* eg. just "Ni330", or "aster" */
	}
	
	if (dh&1) { CL_ERROR(); }
	
	/* NES ROM header loaded successfully */
	cart->crc32=crc;
	cart->crc32_patch=crc_p;
	strcpy(cart->name,file->name);
	strcpy(cart->filename,file->filename);
	cart->prgbanks_real=(header[4]<<1);
	cart->vrombanks_real=(header[5]<<3);
	ines=(header[7]&0xf0)|(header[6]>>4);
	cart->mirroring=((header[6]&1)?CARTRIDGE_MIRRORING_VERTICAL:CARTRIDGE_MIRRORING_HORIZONTAL)|(((header[6]>>3)&1)?CARTRIDGE_MIRRORING_4:0);
	cart->battery=(header[6]>>1)&1;
	cart->mode=(header[9]&1)&(header[9]<2);
	cart->trainer=(header[6]>>2)&1;
	cart->type=mapperinfo_ines2maptype(ines);
	
	/* incorrect bank size(s) */
	if (((cart->prgbanks_real<<13)+(cart->vrombanks_real<<10))<(file->size&0xffffe000)) {
		int fs=file->size&0xffffe000,pb,vb;
		for (pb=cart->prgbanks_real<<13;pb<=fs;pb<<=1) {
			for (vb=cart->vrombanks_real<<10;vb&&(pb+vb)<fs;vb<<=1) { ; }
			if ((pb+vb)==fs) {
				pb>>=13; vb>>=10;
				LOG(LOG_MISC|LOG_WARNING,"suggest banks change: %d/%d to %d/%d\n",cart->prgbanks_real<<3,cart->vrombanks_real,pb<<3,vb);
				/*cart->prgbanks_real=pb; cart->vrombanks_real=vb;*/ /* don't fix them, there are too many 'overdumps' out there that might cause problems here */
				break;
			}
		}
	}
	
	if (cart->prgbanks_real>CARTRIDGE_MAXPRG) { LOG(LOG_MISC|LOG_WARNING,"PRG banks overflow\n"); CL_ERROR(); }
	if (cart->vrombanks_real>CARTRIDGE_MAXVROM) { LOG(LOG_MISC|LOG_WARNING,"VROM banks overflow\n"); CL_ERROR(); }
	
	cart->prgbanks=cart->prgbanks_real-1;
	cart->vrombanks=cart->vrombanks_real-(cart->vrombanks_real!=0);
	
	/* rom open dialog settings override */
	if (!get&&*s_crc==crc&&*s_crc_p==crc_p) {
		cart->type=*s_maptype;
		cart->mirroring=*s_mir?CARTRIDGE_MIRRORING_VERTICAL:CARTRIDGE_MIRRORING_HORIZONTAL;
		cart->battery=*s_battery;
	}
	
	/* detect mapper type */
	else {
		/* only MMC3 NES-TR1ROM/NES-TVROM, and an Irem one (Napoleon Senki) support 4-screen 'mirroring' */
		if (cart->mirroring&CARTRIDGE_MIRRORING_4) {
			switch (cart->type) {
				case MAPTYPE_NINTENDO_MMC3_TXROM: cart->type=MAPTYPE_NINTENDO_MMC3_TR1ROM_TVROM; break;
				case MAPTYPE_IREM_SIMPLE_M4: break;
				
				default:
					cart->mirroring=CARTRIDGE_MIRRORING_VERTICAL;
					LOG(LOG_MISC|LOG_WARNING,"M-4 is unsupported, guessed to M-V\n");
					break;
			}
		}
		
		switch (cart->type) {
			/* MMC1 with 2/4 pages of ram, detect as SOROM/SXROM */
			case MAPTYPE_NINTENDO_MMC1_SxROM:
				if (header[8]==2) cart->type=MAPTYPE_NINTENDO_MMC1_SOROM;
				else if (header[8]==4) cart->type=MAPTYPE_NINTENDO_MMC1_SXROM;
				break;
			
			/* MMC3 with battery, do MMC6 check after file close */
			case MAPTYPE_NINTENDO_MMC3_TXROM:
				mmc6check=cart->battery;
				break;
			
			/* BNROM and NINA-001 share the same iNES mapper number, if it contains CHR ROM, detect as NINA-001 */
			case MAPTYPE_NINTENDO_BXROM:
				if (cart->vrombanks) cart->type=MAPTYPE_AVE_NINA_001;
				break;
			
			/* mapper type out of spec bounds, try to guess (only for most common types) */
			case MAPTYPE_NINTENDO_NROM:
				if (cart->vrombanks&&cart->prgbanks>3) { cart->type=MAPTYPE_NINTENDO_MMC3_TXROM; LOG(LOG_MISC|LOG_WARNING,"NROM is incorrect, guessed to MMC3\n"); }
				else if (cart->prgbanks>3) { cart->type=MAPTYPE_NINTENDO_UXROM; LOG(LOG_MISC|LOG_WARNING,"NROM is incorrect, guessed to UxROM\n"); }
				else if (cart->vrombanks>7) { cart->type=MAPTYPE_NINTENDO_CNROM; LOG(LOG_MISC|LOG_WARNING,"NROM is incorrect, guessed to CNROM\n"); }
				break;
			case MAPTYPE_NINTENDO_CNROM:
				if (cart->vrombanks&&cart->prgbanks>3) { cart->type=MAPTYPE_NINTENDO_MMC3_TXROM; LOG(LOG_MISC|LOG_WARNING,"CNROM is incorrect, guessed to MMC3\n"); }
				else if (cart->prgbanks>3) { cart->type=MAPTYPE_NINTENDO_UXROM; LOG(LOG_MISC|LOG_WARNING,"CNROM is incorrect, guessed to UxROM\n"); }
				break;
			case MAPTYPE_NINTENDO_UXROM:
				if (cart->vrombanks>7&&cart->prgbanks<4) { cart->type=MAPTYPE_NINTENDO_CNROM; LOG(LOG_MISC|LOG_WARNING,"UxROM is incorrect, guessed to CNROM\n"); }
				else if (cart->vrombanks>7) { cart->type=MAPTYPE_NINTENDO_MMC3_TXROM; LOG(LOG_MISC|LOG_WARNING,"UxROM is incorrect, guessed to MMC3\n"); }
				break;
			
			default: break;
		}
	}
	
	/* wants header info */
	if (get) {
		int old_zip=file->is_zip; /* gets destroyed in cartridge_check_mmc6, needed in nes.c */
		file_patch_close(); file_close();
		if (mmc6check) cartridge_check_mmc6(cart);
		*s_maptype=cart->type;
		*s_battery=cart->battery;
		*s_mir=cart->mirroring>>1&1; /* 1: v, 0: h */
		*s_crc=crc; *s_crc_p=crc_p;
		file->is_zip=old_zip; return TRUE;
	}
	
	if (!mapperinfo_supported(cart->type)) {
		if (mapperinfo_ines2maptype(ines)!=MAPTYPE_UNKNOWN) LOG(LOG_MISC|LOG_WARNING,"can't open with unknown mapper type\n");
		else LOG(LOG_MISC|LOG_WARNING,"iNES mapper %d is unsupported\n",ines);
		CL_ERROR();
	}
	
	/* trainer should probably be enabled */
	if (!cart->trainer&&((cart->prgbanks_real<<13)+(cart->vrombanks_real<<10)+0x210)==file->size&&(cart->type==MAPTYPE_FRONT_FAREAST_F3XXX||cart->type==MAPTYPE_FRONT_FAREAST_F4XXX||cart->type==MAPTYPE_FRONT_FAREAST_F8XXX)) {
		cart->trainer=TRUE;
		LOG(LOG_MISC|LOG_WARNING,"trainer flag contradicted file size\n");
	}
	
	
	/* allocate wram */
	MEM_CREATE(cart->wram,0x2000); memset(cart->wram,0xff,0x2000);
	
	/* load trainer */
	if (cart->trainer) {
		if (!file_read(cart->wram+0x1000,0x200)) { LOG(LOG_MISC|LOG_WARNING,"ROM read error at trainer\n"); CL_ERROR(); }
		if (CARTRIDGE_SKIP_TRAINER) {
			memset(cart->wram,0xff,0x2000);
			cart->trainer=FALSE;
		}
		else { LOG(LOG_VERBOSE,"trainer loaded\n"); }
	}
	
	/* load prg banks */
	MEM_CREATE(cart->dummy_rom,0x2000);
	for (i=0;i<CARTRIDGE_MAXPRG;i++) cart->prg[i]=cart->dummy_rom;
	
	LOG(LOG_VERBOSE,"loading %dK PRG-ROM",cart->prgbanks_real<<3);
	for (i=0;i<=cart->prgbanks;i++) {
		MEM_CREATE(cart->prg[i],0x2000);
		LOG(LOG_VERBOSE,".");
		if (!file_read(cart->prg[i],0x2000)) { LOG(LOG_MISC|LOG_WARNING,"ROM read error at prg bank %d\n",i); CL_ERROR(); }
	}
	LOG(LOG_VERBOSE,"done\n");
	
	/* load vrom banks */
	MEM_CREATE(cart->dummy_vrom,0x400);
	for (i=0;i<CARTRIDGE_MAXVROM;i++) cart->vrom[i]=cart->dummy_vrom;
	
	if (cart->vrombanks) {
		LOG(LOG_VERBOSE,"loading %dK VROM",cart->vrombanks_real);
		for (i=0;i<=cart->vrombanks;i++) {
			MEM_CREATE(cart->vrom[i],0x400);
			if (((i&0xf)|8)==8) LOG(LOG_VERBOSE,".");
			if (!file_read(cart->vrom[i],0x400)) { LOG(LOG_MISC|LOG_WARNING,"ROM read error at chr bank %d\n",i); CL_ERROR(); }
		}
		LOG(LOG_VERBOSE,"done\n");
	}
	else for (i=0;i<8;i++) { MEM_CREATE(cart->vram[i],0x400); } /* allocate vram bank */
	
	
	/* auto detect region */
	cart->region=CARTRIDGE_REGION_USA;
	if (strlen(cart->name)>3)
		for (i=0;i<strlen(cart->name)-3;i++) {
			if (memcmp(cart->name+i," (J)",4)==0) { cart->region=CARTRIDGE_REGION_JAPAN; break; }
			else if ((memcmp(cart->name+i," (E)",4)==0)/*||(memcmp(cart->name+i," (As)",5)==0)*/) { cart->mode=TRUE; break; } /* As=GoodNES Asia, should be PAL, but this breaks lots of games, was the NES officially released over there? */
		}
	if (cart->mode) cart->region=CARTRIDGE_REGION_EUROPE;
	
	/* wrong prg/vrom size? */
	#define UNEVEN(w,s) \
		for (i=0;cart->w##banks>>i;i++) { ; } \
		\
		if (i>1) { \
			u16 j=0,banks=0; \
			banks=BIT(i)-1; \
			i--; \
			while (i--) if ((cart->w##banks&BIT(i))==0) j=TRUE; \
			if (j) { \
				LOG(LOG_MISC|LOG_WARNING,"uneven %s banks: %%",s); \
				LOG_BIN(LOG_MISC|LOG_WARNING,cart->w##banks,TRUE); \
				\
				i=banks; j=cart->w##banks; \
				while (i>cart->w##banks) { \
					cart->w[i]=cart->w[j&banks]; \
					i--; j--; \
				} \
				cart->w##banks=banks; \
			} \
		}
	UNEVEN(prg,"prg")
	UNEVEN(vrom,"chr")
	
	/* check loaded data size vs file size */
	i=(cart->prgbanks_real<<13)+(cart->vrombanks_real<<10)+(cart->trainer*0x200)+0x10;
	if (file->size!=i) LOG(LOG_MISC|LOG_WARNING,"suggest size change: $%X to $%X\n",file->size,i);
	
	LOG(LOG_VERBOSE,"loaded '%s', %s region\n",cart->name,cart->region==CARTRIDGE_REGION_USA?"NTSC-U":cart->region==CARTRIDGE_REGION_JAPAN?"NTSC-J":cart->region==CARTRIDGE_REGION_EUROPE?"PAL":"unknown");
	
	file_patch_close(); file_close();
	if (mmc6check) cartridge_check_mmc6(cart);
	return TRUE;
}

void cartridge_unload(Cartridge* cart)
{
	int i;
	for (i=0;i<cart->prgbanks_real;i++) { if (cart->prg[i]!=cart->dummy_rom) { MEM_CLEAN(cart->prg[i]); } }
	for (i=0;i<cart->vrombanks_real;i++) { if (cart->vrom[i]!=cart->dummy_vrom) { MEM_CLEAN(cart->vrom[i]); } }
	for (i=0;i<CARTRIDGE_MAXVRAM;i++) { MEM_CLEAN(cart->vram[i]); }
	
	MEM_CLEAN(cart->dummy_rom);
	MEM_CLEAN(cart->dummy_vrom);
	MEM_CLEAN(cart->wram);
	
	memset(cart,0,sizeof(Cartridge));
	
	LOG(LOG_VERBOSE,"unloaded cartridge\n");
}


void cartridge_battery_load(void)
{
	if (cartridge->battery) {
		int size=cartridge->battery_size;
		u8* data=cartridge->battery_data;
		if (data==NULL) data=cartridge->cur_wrambank;
		
		file_setfile(&file->batterydir,cartridge->name,"sav",NULL);
		if (file_open()) {
			if (size!=file->size) {
				if (size>file->size) size=file->size;
				LOG(LOG_MISC|LOG_WARNING,"battery backed SRAM sizes mismatch\n"); /* no problem for MMC5 */
			}
			if (file_read(data,size)) LOG(LOG_VERBOSE,"loaded battery backed SRAM\n");
			else LOG(LOG_MISC|LOG_WARNING,"battery backed SRAM read error\n");
		}
		else {
			if (file_accessible()) LOG(LOG_MISC|LOG_WARNING,"couldn't load battery backed SRAM\n");
			else {
				u8 temp=0;
				file_close();
				if (!file_save()||!file_write(&temp,0)) LOG(LOG_MISC|LOG_WARNING,"couldn't initialize battery backed SRAM\n");
			}
		}
		file_close();
		mapper_battery_custom_load();
	}
}

void cartridge_battery_save(int h)
{
	if (cartridge->battery&&!h&&!mapper_battery_custom_save()) {
		u8* data=cartridge->battery_data;
		if (data==NULL) data=cartridge->cur_wrambank;
		
		file_setfile(&file->batterydir,cartridge->name,"sav",NULL);
		if (file_save()) {
			if (file_write(data,cartridge->battery_size)) LOG(LOG_VERBOSE,"saved battery backed SRAM\n");
			else LOG(LOG_MISC|LOG_WARNING,"battery backed SRAM write error\n");
		}
		else LOG(LOG_MISC|LOG_WARNING,"couldn't save battery backed SRAM\n");
		file_close();
	}
}


void cartridge_init(void)
{
	MEM_CREATE(cartridge,sizeof(Cartridge));
	MEM_CREATE(cartridge_temp,sizeof(Cartridge));
	
	LOG(LOG_VERBOSE,"cartridge initialised\n");
}

void cartridge_clean(void)
{
	cartridge_unload(cartridge);
	MEM_CLEAN(cartridge);
	MEM_CLEAN(cartridge_temp);
	
	LOG(LOG_VERBOSE,"cartridge cleaned\n");
}

void cartridge_set_cartridge(void)
{
	Cartridge* cart;
	cart=cartridge;
	cartridge=cartridge_temp;
	cartridge_temp=cart;
}
