#include "global.h"
#include "cartridge.h"
#include "ppu_memorymap.h"
#include "ppu.h"

/* memory map

	$10000
	$c000	mirror of $0000-$3fff/not accessible(addressbus=15 bits)?
	$8000	mirror of $0000-$3fff/not accessible(addressbus=15 bits)?
	$4000	mirror of $0000-$3fff
	$3f20	mirror of palette
	$3f10	sprite palette
	$3f00	tiles palette
	$3000	mirror of $2000-$2eff
	$2fc0	attribute table 3
	$2c00	name table 3
	$2bc0	attribute table 2
	$2800	name table 2
	$27c0	attribute table 1
	$2400	name table 1
	$23c0	attribute table 0
	$2000	name table 0
	$1000	pattern table 1
	$0000	pattern table 0
*/

static struct {
	u8 palette[0x20];
	
	u8* name_ram[2];
	u8* cur_name[4];
	
	u8 buffer;
	
	u8* ppu_control2_ptr;
} ppu_memmap;

void __fastcall (*ppubusoverride)(u16);

u8* ppu_memmap_get_buffer_ptr(void) { return &ppu_memmap.buffer; }
u8* ppu_memmap_get_palette_ptr(void) { return ppu_memmap.palette; }
u8** ppu_memmap_get_cur_name_ptr(void) { return ppu_memmap.cur_name; }
u8** ppu_memmap_get_name_ram_ptr(void) { return ppu_memmap.name_ram; }

void ppu_memmap_set_ppubusoverride(fp_ppubusoverride fun) { ppubusoverride=fun; }

void ppu_memmap_init(void)
{
	/* palette values on startup: probably different on each NES, this is blargg's NES */
	/*
	const u8 palette_startup[]={
		0x09,0x01,0x00,0x01,0x00,0x02,0x02,0x0d,0x08,0x10,0x08,0x24,0x00,0x00,0x04,0x2c,
		0x09,0x01,0x34,0x03,0x00,0x04,0x00,0x14,0x08,0x3a,0x00,0x02,0x00,0x20,0x2c,0x08
	};
	*/
	
	memset(&ppu_memmap,0,sizeof(ppu_memmap));
	
	/*memcpy(ppu_memmap.palette,palette_startup,0x20);*/
	memset(ppu_memmap.palette,BIN8(00111111),0x20);
	
	MEM_CREATE(ppu_memmap.name_ram[0],0x400);
	MEM_CREATE(ppu_memmap.name_ram[1],0x400);
	ppu_memmap.ppu_control2_ptr=ppu_get_control2_ptr();
	
	ppu_set_fpt(PPU_FPT_NORMAL,FALSE,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL);
	
	LOG(LOG_VERBOSE,"PPU memorymap initialised\n");
}

void ppu_memmap_clean(void)
{
	MEM_CLEAN(ppu_memmap.name_ram[0]);
	MEM_CLEAN(ppu_memmap.name_ram[1]);
	
	LOG(LOG_VERBOSE,"PPU memorymap cleaned\n");
}

void ppu_memmap_printf_internal_nametable(int n)
{
	/* debug related */
	int i;
	for (i=0;i<0x400;i++) {
		printf("0x%02X",ppu_memmap.name_ram[0][i]);
		if (i!=0x3ff) printf(", ");
	}
	printf("\n");
}


/* --- pattern tables */
#define GET_PATDATA_START(x)	u8 ret=ppu_memmap.buffer; ppu_memmap.buffer=cartridge->cur_patbank[x][address&0x3ff]
#define GET_PATDATA_END()	return ret
#define GET_PATDATA(x)		GET_PATDATA_START(x); GET_PATDATA_END();
#define SET_PATDATA(x)		if (cartridge->cur_patbank_isram[x]) cartridge->cur_patbank[x][address&0x3ff]=data

/* read */
/* PT0 */
u8 __fastcall ppu_read_pt0a(u16 address) { GET_PATDATA(0); } /* $0000-$03ff */
u8 __fastcall ppu_read_pt0b(u16 address) { GET_PATDATA(1); } /* $0400-$07ff */
u8 __fastcall ppu_read_pt0c(u16 address) { GET_PATDATA(2); } /* $0800-$0bff */
u8 __fastcall ppu_read_pt0d(u16 address) { GET_PATDATA(3); } /* $0c00-$0fff */
/* PT1 */
u8 __fastcall ppu_read_pt1a(u16 address) { GET_PATDATA(4); } /* $1000-$13ff */
u8 __fastcall ppu_read_pt1b(u16 address) { GET_PATDATA(5); } /* $1400-$17ff */
u8 __fastcall ppu_read_pt1c(u16 address) { GET_PATDATA(6); } /* $1800-$1bff */
u8 __fastcall ppu_read_pt1d(u16 address) { GET_PATDATA(7); } /* $1c00-$1fff */

/* write */
/* PT0 */
void __fastcall ppu_write_pt0a(u16 address,u8 data) { SET_PATDATA(0); } /* $0000-$03ff */
void __fastcall ppu_write_pt0b(u16 address,u8 data) { SET_PATDATA(1); } /* $0400-$07ff */
void __fastcall ppu_write_pt0c(u16 address,u8 data) { SET_PATDATA(2); } /* $0800-$0bff */
void __fastcall ppu_write_pt0d(u16 address,u8 data) { SET_PATDATA(3); } /* $0c00-$0fff */
/* PT1 */
void __fastcall ppu_write_pt1a(u16 address,u8 data) { SET_PATDATA(4); } /* $1000-$13ff */
void __fastcall ppu_write_pt1b(u16 address,u8 data) { SET_PATDATA(5); } /* $1400-$17ff */
void __fastcall ppu_write_pt1c(u16 address,u8 data) { SET_PATDATA(6); } /* $1800-$1bff */
void __fastcall ppu_write_pt1d(u16 address,u8 data) { SET_PATDATA(7); } /* $1c00-$1fff */

/* overrides */
u8 __fastcall ppu_read_pt0a_override(u16 address) { GET_PATDATA_START(0); (*ppubusoverride)(address); GET_PATDATA_END(); }
u8 __fastcall ppu_read_pt0b_override(u16 address) { GET_PATDATA_START(1); (*ppubusoverride)(address); GET_PATDATA_END(); }
u8 __fastcall ppu_read_pt0c_override(u16 address) { GET_PATDATA_START(2); (*ppubusoverride)(address); GET_PATDATA_END(); }
u8 __fastcall ppu_read_pt0d_override(u16 address) { GET_PATDATA_START(3); (*ppubusoverride)(address); GET_PATDATA_END(); }
u8 __fastcall ppu_read_pt1a_override(u16 address) { GET_PATDATA_START(4); (*ppubusoverride)(address); GET_PATDATA_END(); }
u8 __fastcall ppu_read_pt1b_override(u16 address) { GET_PATDATA_START(5); (*ppubusoverride)(address); GET_PATDATA_END(); }
u8 __fastcall ppu_read_pt1c_override(u16 address) { GET_PATDATA_START(6); (*ppubusoverride)(address); GET_PATDATA_END(); }
u8 __fastcall ppu_read_pt1d_override(u16 address) { GET_PATDATA_START(7); (*ppubusoverride)(address); GET_PATDATA_END(); }
void __fastcall ppu_write_pt0a_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(0); }
void __fastcall ppu_write_pt0b_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(1); }
void __fastcall ppu_write_pt0c_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(2); }
void __fastcall ppu_write_pt0d_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(3); }
void __fastcall ppu_write_pt1a_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(4); }
void __fastcall ppu_write_pt1b_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(5); }
void __fastcall ppu_write_pt1c_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(6); }
void __fastcall ppu_write_pt1d_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PATDATA(7); }

/* --- nametables --- */
#define GET_NAMEDATA_START(x)	u8 ret=ppu_memmap.buffer; ppu_memmap.buffer=ppu_memmap.cur_name[x][address&0x3ff]
#define GET_NAMEDATA_END()	return ret
#define GET_NAMEDATA(x)		GET_NAMEDATA_START(x); GET_NAMEDATA_END();
#define SET_NAMEDATA(x)		ppu_memmap.cur_name[x][address&0x3ff]=data

/* read */
u8 __fastcall ppu_read_name0(u16 address) { GET_NAMEDATA(0); } /* nametable 0: $2000-$23ff */
u8 __fastcall ppu_read_name1(u16 address) { GET_NAMEDATA(1); } /* nametable 1: $2400-$27ff */
u8 __fastcall ppu_read_name2(u16 address) { GET_NAMEDATA(2); } /* nametable 2: $2800-$2bff */
u8 __fastcall ppu_read_name3(u16 address) { GET_NAMEDATA(3); } /* nametable 3: $2c00-$2fff */

/* write */
void __fastcall ppu_write_name0(u16 address,u8 data) { SET_NAMEDATA(0); } /* nametable 0: $2000-$23ff */
void __fastcall ppu_write_name1(u16 address,u8 data) { SET_NAMEDATA(1); } /* nametable 1: $2400-$27ff */
void __fastcall ppu_write_name2(u16 address,u8 data) { SET_NAMEDATA(2); } /* nametable 2: $2800-$2bff */
void __fastcall ppu_write_name3(u16 address,u8 data) { SET_NAMEDATA(3); } /* nametable 3: $2c00-$2fff */

/* overrides */
u8 __fastcall ppu_read_name0_override(u16 address) { GET_NAMEDATA_START(0); (*ppubusoverride)(address); GET_NAMEDATA_END(); }
u8 __fastcall ppu_read_name1_override(u16 address) { GET_NAMEDATA_START(1); (*ppubusoverride)(address); GET_NAMEDATA_END(); }
u8 __fastcall ppu_read_name2_override(u16 address) { GET_NAMEDATA_START(2); (*ppubusoverride)(address); GET_NAMEDATA_END(); }
u8 __fastcall ppu_read_name3_override(u16 address) { GET_NAMEDATA_START(3); (*ppubusoverride)(address); GET_NAMEDATA_END(); }
void __fastcall ppu_write_name0_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_NAMEDATA(0); }
void __fastcall ppu_write_name1_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_NAMEDATA(1); }
void __fastcall ppu_write_name2_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_NAMEDATA(2); }
void __fastcall ppu_write_name3_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_NAMEDATA(3); }

/* --- palette --- */
#define GET_PALETTE() \
	ppu_memmap.buffer=ppu_memmap.cur_name[3][address&0x3ff]; \
	if (*ppu_memmap.ppu_control2_ptr&1) return ppu_memmap.palette[address&0x1f]&BIN8(11110000); /* mono */ \
	else return ppu_memmap.palette[address&0x1f]
#define SET_PALETTE() \
	if((address&BIN8(00000011))==0) ppu_memmap.palette[(address&0x1f)^0x10]=data&BIN8(00111111); \
	ppu_memmap.palette[address&0x1f]=data&BIN8(00111111)

u8 __fastcall ppu_read_palette(u16 address) { GET_PALETTE(); }
void __fastcall ppu_write_palette(u16 address,u8 data) { SET_PALETTE(); }

/* overrides */
u8 __fastcall ppu_read_palette_override(u16 address) { (*ppubusoverride)(address); GET_PALETTE(); }
void __fastcall ppu_write_palette_override(u16 address,u8 data) { (*ppubusoverride)(address); SET_PALETTE(); }
